from typing import List, Optional, Tuple
from datetime import timedelta

import bugzoo
from bugzoo.localization import SuspiciousnessMetric

from .problem import Problem
from .candidate import Candidate
from .searcher import Searcher
from .generator import DeletionGenerator


__ALL__ = ['RepairReport', 'repair']


class RepairReport(object):
    def __init__(self,
                 num_candidate_evals: int,
                 num_test_evals: int,
                 duration: timedelta):
        self.__num_candidate_evals = None
        self.__num_test_evals = None
        self.__duration = None

    @property
    def num_candidate_evals(self) -> int:
        """
        Number of candidate evaluations performed during the search.
        """
        return self.__num_candidate_evals

    @property
    def num_test_evals(self) -> int:
        """
        Number of test-case evaluations performed during the search.
        """
        return self.__num_test_evals


    @property
    def duration(self) -> timedelta:
        """
        The wall-clock duration of the search process.
        """
        return self.__duration


def repair(bugzoo: bugzoo.BugZoo,
           problem: Problem,
           seed: Optional[int] = None,
           threads: Optional[int] = 1,
           terminate_early: Optional[bool] = True,
           time_limit: Optional[timedelta] = None
           ) -> Tuple[List[Candidate], RepairReport]:
    """
    Attempts to repair a given program.

    Returns:
        A tuple of the form `(patches, report)`, where `patches` is a list
        of all the candidate patches discovered during the search, and
        `report` contains a summary of the search process.
    """
    # generate the search space
    lines = problem.lines
    candidates = DeletionGenerator(lines)

    searcher = Searcher(bugzoo,
                        problem,
                        candidates,
                        threads=threads,
                        time_limit=time_limit)
    repairs = list(searcher)
    report = RepairReport(searcher.num_candidate_evals,
                          searcher.num_test_evals,
                          searcher.time_running)
    return (repairs, report)



from typing import Iterable, Iterator, Optional
from timeit import default_timer as timer
import datetime
import threading

import bugzoo

from .candidate import Candidate
from .problem import Problem


__ALL__ = ['Searcher']


class Searcher(object):
    def __init__(self,
                 bugzoo: bugzoo.BugZoo,
                 problem: Problem,
                 candidates: Iterable[Candidate],
                 *,
                 threads: int = 1,
                 time_limit: Optional[datetime.timedelta] = None
                 ) -> None:
        """
        Constructs a new searcher for a given source of candidate patches.

        Parameters:
            problem: a description of the problem.
            candidates: a source of candidate patches.
            threads: the number of threads that should be made available to
                the search process.
            time_limit: an optional limit on the amount of time given to the
                searcher.
        """
        assert time_limit is None or time_limit > 0, \
            "if specified, time limit should be greater than zero."

        self.__bugzoo = bugzoo
        self.__problem = problem
        self.__candidates = candidates
        self.__time_limit = time_limit
        self.__num_threads = threads

        # records the time at which the current iteration begun
        self.__time_iteration_begun = None

        self.__lock_candidates = threading.Lock() # type: threading.Lock
        self.__counter_candidates = 0
        self.__counter_tests = 0
        self.__exhausted_candidates = False
        self.__time_running = datetime.timedelta()
        self.__error_occurred = False
        self.__next_patch = None # type: Optional[Candidate]

    @property
    def paused(self) -> bool:
        """
        Indicates whether this searcher is paused.
        """
        return (self.__next_patch is not None) or self.exhausted

    @property
    def exhausted(self) -> bool:
        """
        Indicates whether or not the resources available to this searcher have
        been exhausted.
        """
        if self.__error_occurred:
            return True
        if self.__exhausted_candidates:
            return True
        if self.__time_limit is None:
            return False
        return self.time_running > self.time_limit

    @property
    def num_test_evals(self) -> int:
        """
        The number of test case evaluations that have been performed during
        this search process.
        """
        return self.__counter_tests

    @property
    def num_candidate_evals(self) -> int:
        """
        The number of candidate patches that have been evaluated over the
        course of this search process.
        """
        return self.__counter_candidates

    @property
    def time_limit(self) -> Optional[datetime.timedelta]:
        """
        An optional limit on the length of time that may be spent searching
        for patches.
        """
        return self.__time_limit

    @property
    def time_running(self) -> datetime.timedelta:
        """
        The amount of time that has been spent searching for patches.
        """
        duration_iteration = \
            datetime.timedelta(timer() - self.__time_iteration_begun)
        return self.__time_running + duration_iteration

    def __iter__(self) -> Iterator[Candidate]:
        return self

    def __next__(self) -> Candidate:
        """
        Searches for the next acceptable patch.

        Returns:
            the next patch that passes all tests.

        Raises:
            StopIteration: if the search space or available resources have
                been exhausted.
        """
        self.__time_iteration_begun = timer()
        threads = [] # type: List[threading.Thread]

        # TODO there's a bit of a bug: any patches # type: List[threading.Thread] # type: List[threading.Thread] # type: List[threading.Thread] that were read from the
        #   generator by the worker and were still stored in its
        #   `candidate` variable will be discarded. fixable by adding a
        #   buffer to `_try_next`.
        try:
            def worker(searcher: 'Searcher') -> None:
                while True:
                    if not searcher._try_next():
                        break

            for _ in range(self.__num_threads):
                t = threading.Thread(target=worker, args=(self,), daemon=True)
                threads.append(t)
                t.start()
        except:
            self.__error_occurred = True
        finally:
            # TODO this is bad -- use while instead
            for t in threads:
                t.join()

        duration_iteration = timer() - self.__time_iteration_begun
        self.__time_running += datetime.timedelta(seconds=duration_iteration)

        # if we have a patch, return it
        if self.__next_patch:
            self.__next_patch = None
            return self.__next_patch

        # if not, we're done
        raise StopIteration

    def _try_next(self) -> bool:
        """
        Evaluates the next candidate patch.

        Returns:
            a boolean indicating whether the calling thread should continue to
            evaluate candidate patches.
        """
        if self.paused:
            return False

        self.__lock_candidates.acquire()
        try:
            candidate = next(self.__candidates)
        except StopIteration:
            print("exhausted all candidate patches!")
            self.__exhausted_candidates = True
            return False
        finally:
            self.__lock_candidates.release()

        self.__counter_candidates += 1
        bz = self.__bugzoo
        container = bz.containers.provision(self.__problem.bug)
        try:
            patch = candidate.diff(self.__problem)
            diff = candidate.diff(self.__problem)
            print("Evaluating: {}\n{}\n".format(candidate, diff))
            bz.containers.patch(container, patch)

            # ensure that the patch compiles
            if not bz.containers.compile(container).successful:
                print("Failed to compile: {}".format(candidate))
                return True

            # for now, execute all tests in no particular order
            # TODO perform test ordering
            for test in self.__problem.tests:
                print("Executing test: {} ({})".format(test.name, candidate))
                self.__counter_tests += 1
                outcome = bz.containers.execute(container, test)
                if not outcome.passed:
                    print("Failed test: {} ({})".format(test.name, candidate))
                    return True
                print("Passed test: {} ({})".format(test.name, candidate))

            # FIXME possible race condition if two workers report repairs at
            #   the same time?
            # if we've found a repair, pause the search
            self.__next_patch = candidate

            # TODO make this prettier
            # report the patch
            time_repair = self.time_running.seconds / 60.0
            msg = "FOUND REPAIR [{:.2f} minutes]: {}\n{}\n{}\n{}"
            msg = msg.format(time_repair, candidate, ("=" * 80), diff, ("="*80))
            print(msg)

            return True

        # TODO ensure a bool is returned when an exception occurs
        finally:
            print("Evaluated: {}".format(candidate))
            if container:
                del bz.containers[container.uid]



from bugzoo.core.fileline import FileLine


def ends_with_semi_colon(contents: str) -> bool:
    """
    Returns `True` if a given line ends with a semi-colon.
    """
    return contents.rstrip().endswith(';')


def has_balanced_delimiters(contents: str) -> bool:
    """
    Checks whether all delimiters that occur in a given snippet are balanced.
    """
    return True



from typing import List, Iterator, Dict

from bugzoo.core.bug import Bug
from bugzoo.core.coverage import FileLine

from darjeeling.donor import Snippet, DonorPool
from darjeeling.source import SourceFile


class Transformation(object):
    @property
    def diff(self) -> str:
        raise NotImplementedError


class DeleteTransformation(object):
    def __init__(self, line: FileLine) -> None:
        self.__line = line

    @property
    def line(self) -> FileLine:
        return self.__line

    def __str__(self) -> str:
        return "DELETE[{}]".format(self.__line)


class ReplaceTransformation(object):
    """
    Replaces a numbered line in a given file with a provided snippet.
    """
    def __init__(self, line: FileLine, snippet: Snippet) -> None:
        self.__line = line
        self.__snippet = snippet

    @property
    def line(self) -> FileLine:
        return self.__line

    @property
    def snippet(self) -> Snippet:
        return self.__snippet

    def __str__(self) -> str:
        return "REPLACE[{}; {}]".format(self.__line, self.__snippet)


class AppendTransformation(object):
    """
    Appends a given snippet to a specific line in a given file.
    """
    def __init__(self, line: FileLine, snippet: Snippet) -> None:
        self.__line = line
        self.__snippet = snippet

    @property
    def snippet(self) -> Snippet:
        return self.__snippet

    @property
    def line(self) -> FileLine:
        return self.__line

    def __str__(self) -> str:
        return "APPEND[{}; {}]".format(self.__line, self.__snippet)


class TransformationDatabase(object):
    """
    Used to represent the space of all atomic program transformations.
    """
    @staticmethod
    def generate(bug: Bug,
                 pool: DonorPool,
                 sources: Dict[str, SourceFile],
                 lines: List[FileLine]) -> None:
        transformations = []
        for line in lines:
            # get the content of the line
            original_line = sources[line.filename][line.num].strip()

            # ignore any line that might not be a statement
            if not original_line.endswith(';'):
                continue

            # ignore comments and macros (probably redundant)
            if original_line.startswith('#') or \
               original_line.startswith('//') or \
               original_line.startswith('/*') \
            :
               continue

            # deletion
            transformations.append(DeleteTransformation(line))

            # replace and append
            for snippet in pool:
                transformations.append(AppendTransformation(line, snippet))

                # don't replace a line with an equivalent one
                if snippet.content != original_line:
                    transformations.append(ReplaceTransformation(line, snippet))

        return TransformationDatabase(transformations)

    def __init__(self, transformations: List[Transformation]):
        self.__transformations = transformations[:]

    def __len__(self) -> int:
        return len(self.__transformations)

    def __iter__(self) -> Iterator[Transformation]:
        """
        Returns an iterator over the transformations contained within this
        database.
        """
        for t in self.__transformations:
            yield t

    def at_line(self, line: FileLine) -> Iterator[Transformation]:
        """
        Returns an iteratover over all the transformations at a given line.
        """
        raise NotImplementedError



from typing import Iterator, List, Iterable

from bugzoo.core.fileline import FileLine

from .candidate import Candidate
from .transformation import Transformation, \
                            AppendTransformation, \
                            DeleteTransformation, \
                            ReplaceTransformation


class CandidateGenerator(object):
    """
    Candidate generators are used to generate (normally in a lazy fashion) a
    stream of candidate patches. For now, candidate generators implement
    one-way communications: they do not accept inputs.
    """
    def __init__(self) -> Iterator[Candidate]:
        return self

    def __next__(self) -> Candidate:
        raise NotImplementedError


class DeletionGenerator(CandidateGenerator):
    def __init__(self, lines: Iterable[FileLine]) -> None:
        """
        Constructs a deletion generator.

        Parameters:
            lines: a sequence of lines for which deletion transformations
                should be generated.
        """
        self.__lines = reversed(list(lines))

    def __next__(self) -> Candidate:
        """
        Returns the next deletion transformation from this generator.
        """
        try:
            next_line = next(self.__lines)
        except StopIteration:
            raise StopIteration

        # TODO add static analysis
        # should we delete this line?
        # * don't delete declarations

        transformation = DeleteTransformation(next_line)
        return Candidate([transformation])


class ReplacementGenerator(CandidateGenerator):
    def __init__(self, lines: Iterable[FileLine]) -> None:
        pass


class AppendGenerator(CandidateGenerator):
    def __init__(self) -> None:
        pass



from typing import List, Iterator

from bugzoo.core.patch import Patch

from darjeeling.transformation import Transformation, \
                                      DeleteTransformation, \
                                      ReplaceTransformation, \
                                      AppendTransformation
from darjeeling.problem import Problem


class Candidate(object):
    """
    Represents a candidate repair as a set of atomic program transformations.
    """
    def __init__(self,
                 transformations: List[Transformation]
                 ) -> None:
        self.__transformations = frozenset(transformations)

    def __str__(self) -> str:
        transformations_s = \
            ', '.join([str(t) for t in self.transformations])
        return "<{}>".format(transformations_s)

    @property
    def transformations(self) -> Iterator[Transformation]:
        """
        The transformations that comprise this repair.
        """
        for t in self.__transformations:
            yield t

    def diff(self, problem: Problem) -> Patch:
        """
        Computes a plaintext diff for this repair.
        """
        # group transformations by file
        tf = {}
        for t in self.transformations:
            fn = t.line.filename
            if fn not in tf:
                tf[fn] = []
            tf[fn].append(t)

        # order each group of transformations in descending order of
        # line number
        for group in tf.values():
            group.sort(key=lambda t: t.line.num, reverse=True)

        # transform each group into a modified source code file
        transformed = {}
        for (fn, transformations) in tf.items():
            src = problem.source(fn)
            for t in transformations:
                if isinstance(t, DeleteTransformation):
                    src = src.with_line_removed(t.line.num)
                elif isinstance(t, AppendTransformation):
                    src = src.with_line_inserted(t.line.num + 1, t.snippet.content)
                elif isinstance(t, ReplaceTransformation):
                    src = src.with_line_replaced(t.line.num, t.snippet.content)
                else:
                    raise Exception("unsupported transformation")
            transformed[fn] = src

        # compute a diff for each modified source code file
        diffs = []
        for (fn, source_after) in transformed.items():
            source_before = problem.source(fn)
            diffs.append(source_before.diff(source_after))

        # combine the diffs for each file into a single diff
        diff = '\n'.join(diffs)
        return Patch.from_unidiff(diff)



from typing import List, Optional, Dict, Iterator
import tempfile
import logging

import bugzoo
import bugzoo.localization.suspiciousness as metrics
from bugzoo.core.fileline import FileLine, FileLineSet
from bugzoo.core.bug import Bug
from bugzoo.core.patch import Patch
from bugzoo.core.coverage import TestSuiteCoverage
from bugzoo.core.spectra import Spectra
from bugzoo.localization import SuspiciousnessMetric
from bugzoo.testing import TestCase

import darjeeling.filters as filters
from .donor import DonorPool
from .transformation import TransformationDatabase
from .source import SourceFile


class Problem(object):
    """
    Used to provide a description of a problem (i.e., a bug), and to hold
    information pertinent to its solution (e.g., coverage, transformations).
    """
    def __init__(self,
                 bz: bugzoo.BugZoo,
                 bug: Bug,
                 *,
                 suspiciousness_metric: Optional[SuspiciousnessMetric] = None,
                 in_files: List[str],
                 restrict_to_lines: Optional[FileLineSet] = None,
                 cache_coverage: bool = True,
                 verbose: bool = False
                 ) -> None:
        """
        Constructs a Darjeeling problem description.

        Params:
            bug: A description of the faulty program.
            in_files: An optional list that can be used to restrict the set of
                transformations to those that occur in any files belonging to
                that list. If no list is provided, all source code files will
                be included.
            in_functions: An optional list that can be used to restrict the set
                of transformations to those that occur in any function whose
                name appears in the given list. If no list is provided, no
                filtering of transformations based on the function to which
                they belong will occur.
        """
        assert len(in_files) > 0
        self.__bug = bug
        self.__verbose = verbose

        self.__logger = \
            logging.getLogger('darjeeling.problem').getChild(bug.name)
        # TODO stream logging info for now
        self.__logger.setLevel(logging.DEBUG)
        self.__logger.addHandler(logging.StreamHandler())
        self.__logger.debug("creating problem for bug: %s", bug.name)

        if suspiciousness_metric is None:
            self.__logger.debug("no suspiciousness metric provided: using Tarantula as a default.")
            suspiciousness_metric = metrics.tarantula

        # fetch coverage information
        if cache_coverage:
            self.__logger.debug("fetching coverage information from BugZoo")
            self.__coverage = bz.bugs.coverage(bug)
            self.__logger.debug("fetched coverage information from BugZoo")
        else:
            self.__logger.debug("computing coverage information")
            try:
                container = bz.containers.provision(bug)
                self.__coverage = bz.coverage.coverage(container, bug.tests)
            finally:
                del bz.containers[container.uid]
            self.__logger.debug("computed coverage information")

        # restrict coverage information to specified files
        self.__logger.debug("restricting coverage information to files:\n* %s",
                            '\n* '.join(in_files))
        self.__coverage = self.__coverage.restricted_to_files(in_files)
        self.__logger.debug("restricted coverage information.")

        # determine the passing and failing tests by using coverage information
        self.__logger.debug("using test execution used to generate coverage to determine passing and failing tests")

        self.__tests_failing = set()
        self.__tests_passing = set()
        for test_name in self.__coverage:
            test = bug.harness[test_name]
            test_coverage = self.__coverage[test_name]
            if test_coverage.outcome.passed:
                self.__tests_passing.add(test)
            else:
                self.__tests_failing.add(test)

        # TODO throw an error if there are no failing tests
        self.__logger.info("determined passing and failing tests")
        self.__logger.info("* passing tests: %s", ', '.join([t.name for t in self.__tests_passing]))
        self.__logger.info("* failing tests: %s", ', '.join([t.name for t in self.__tests_failing]))

        # determine the implicated lines
        # 0. we already restricted to lines that occur in specified files
        # 1. restrict to lines covered by failing tests
        # 3. restrict to lines with suspiciousness greater than zero
        # 4. restrict to lines that are optionally provided
        self.__logger.info("Determining implicated lines")
        self.__lines = self.__coverage.failing.lines

        if restrict_to_lines is not None:
            self.__lines = self.__lines.intersection(restrict_to_lines)

        # cache contents of implicated files
        # FIXME this method of construction is slow and error-prone
        self.__logger.debug("storing contents of source code files")
        self.__sources = \
            {fn: SourceFile.load(bz, bug, fn) for fn in self.__lines.files}
        self.__logger.debug("finished storing contents of source code files")

        # restrict attention to statements
        # FIXME for now, we approximate this -- going forward, we can use
        #   Rooibos to determine transformation targets
        line_content_filters = [
            filters.ends_with_semi_colon,
            filters.has_balanced_delimiters
        ]
        for fltr_content in line_content_filters:
            fltr_line = \
                lambda fl: fltr_content(self.__sources[fl.filename][fl.num])
            self.__lines = self.__lines.filter(fltr_line)

        # TODO raise an exception if there are no implicated lines

        # once again, restrict coverage to implicated files to save memory,
        # then generate a spectra and use it to compute the localization
        self.__coverage = \
            self.__coverage.restricted_to_files(self.__lines.files)
        self.__spectra = Spectra.from_coverage(self.__coverage)

        # report implicated lines and files
        self.__logger.info("implicated lines [%d]:\n%s",
                           len(self.__lines), self.__lines)
        self.__logger.info("implicated files [%d]:\n* %s",
                           len(self.__lines.files),
                           '\n* '.join(self.__lines.files))

        # FIXME remove files that are no longer implicated from the cache


    @property
    def bug(self) -> Bug:
        """
        A description of the bug, provided by BugZoo.
        """
        return self.__bug

    @property
    def tests(self) -> Iterator[TestCase]:
        """
        Returns an iterator over the tests for this problem.
        """
        for test in self.__tests_failing:
            yield test
        for test in self.__tests_passing:
            yield test

    @property
    def tests_failing(self) -> Iterator[TestCase]:
        """
        Returns an iterator over the failing tests for this problem.
        """
        return self.__tests_failing.__iter__()

    @property
    def tests_passing(self) -> Iterator[TestCase]:
        """
        Returns an iterator over the passing tests for this problem.
        """
        return self.__tests_passing.__iter__()

    @property
    def coverage(self) -> TestSuiteCoverage:
        """
        Line coverage information for each test within the test suite for the
        program under repair.
        """
        return self.__coverage

    @property
    def spectra(self) -> Spectra:
        """
        Provides a concise summary of the number of passing and failing tests
        that cover each implicated line.
        """
        return self.__spectra

    @property
    def lines(self) -> Iterator[FileLine]:
        """
        Returns an iterator over the lines that are implicated by the
        description of this problem.
        """
        return self.__lines.__iter__()

    def source(self, fn: str) -> SourceFile:
        """
        Returns the contents of a given source code file for this problem.
        """
        return self.__sources[fn]

    def check_sanity(self,
                     expected_to_pass: List[TestCase],
                     expected_to_fail: List[TestCase]
                     ) -> bool:
        """

        Parameters:
            expected_to_pass: a list of test cases for the program that are
                expected to pass.
            expected_to_fail: a list o test cases for the program that are
                expected to fail.

        Returns:
            True if the outcomes of the test executions match those that are
            expected by the parameters to this method.

        Raises:
            errors.UnexpectedTestOutcomes: if the outcomes of the program
                under test, observed while computing coverage information for
                the problem, differ from those that were expected.
        """
        # determine passing and failing tests
        self.__logger.debug("sanity checking...")
        raise NotImplementedError



from typing import List, Iterator
import os
import tempfile
import difflib

import bugzoo
from bugzoo.core.bug import Bug
from bugzoo.core.patch import FilePatch


class SourceFile(object):
    # FIXME refactor
    @staticmethod
    def load(bz: bugzoo.BugZoo, bug: Bug, filename: str) -> 'SourceFile':
        """
        Loads a specified source-code file belonging to a provided faulty
        program version.
        """
        container = bz.containers.provision(bug)
        fd, host_fn = tempfile.mkstemp()
        try:
            os.close(fd)
            container_fn = os.path.join(bug.source_dir, filename)
            bz.containers.copy_from(container, container_fn, host_fn)

            with open(host_fn, 'r') as f:
                lines = [l.rstrip('\n') for l in f]
                return SourceFile(filename, lines)
        finally:
            del bz.containers[container.uid]
            os.remove(host_fn)

    def __init__(self, name: str, lines: List[str]) -> None:
        self.__name = name
        self.__lines = lines[:]

    @property
    def name(self) -> str:
        """
        The name of this source code file.
        """
        return self.__name

    @property
    def lines(self) -> List[str]:
        """
        Returns a copy of the lines contained within this file.
        """
        return self.__lines[:]

    def __getitem__(self, num: int) -> str:
        """
        Retrieves the contents of a given line in this file, specified by its
        one-indexed line number.
        """
        return self.__lines[num - 1]

    def __iter__(self) -> Iterator[str]:
        """
        Returns an iterator over the lines contained within this file.
        """
        for line in self.__lines:
            yield line

    def __len__(self) -> int:
        """
        Returns a count of the number of lines in this file.
        """
        return len(self.__lines)

    def with_line_removed(self, num: int) -> 'SourceFile':
        """
        Returns a variant of this file with a given line, specified by its
        one-indexed line number, removed.
        """
        num -= 1
        l2 = self.__lines[0:num] + self.__lines[num + 1:]
        return SourceFile(self.name, l2)

    def with_line_replaced(self, num: int, replacement: str) -> 'SourceFile':
        """
        Returns a variant of this file with the contents of a given line,
        specified by its one-indexed line number, replaced by a provided string.
        """
        num -= 1
        l2 = self.__lines[:]
        l2[num] = replacement
        return SourceFile(self.name, l2)

    def with_line_inserted(self, num: int, insertion: str) -> 'SourceFile':
        """
        Returns a variant of this file with a given line inserted at a
        specified location.
        """
        num -= 1
        l2 = self.__lines[:]
        l2.insert(num, insertion)
        return SourceFile(self.name, l2)

    def diff(self,
             other: 'SourceFile'
             ) -> str:
        a = ['{}\n'.format(l) for l in self.__lines]
        b = ['{}\n'.format(l) for l in other.lines]
        diff_lines = difflib.unified_diff(a, b,
                                          fromfile=self.name,
                                          tofile=other.name)
        return ''.join(diff_lines)



from typing import List, Iterator, Set, Iterable
from tempfile import NamedTemporaryFile
import os
import tempfile

import bugzoo
from bugzoo.core.bug import Bug
from bugzoo.core.coverage import FileLine

from .source import SourceFile


class Snippet(object):
    """
    Represents a donor code snippet.
    """
    def __init__(self, content: str):
        self.__content = content

    @property
    def content(self) -> str:
        """
        Returns the contents of the snippet as a string.
        """
        return self.__content

    def __str__(self) -> str:
        return self.__content

    def __eq__(self, other) -> bool:
        return isinstance(other, Snippet) and self.content == other.content

    def __hash__(self) -> int:
        return hash(self.__content)


class DonorPool(object):
    @staticmethod
    def from_files(bz: bugzoo.BugZoo,
                   bug: Bug,
                   filenames: List[str]
                   ) -> 'DonorPool':
        file_snippets = [DonorPool.from_file(bz, bug, fn) for fn in filenames]
        all_snippets = [s for pool in file_snippets for s in pool]
        return DonorPool(all_snippets)

    @staticmethod
    def from_file(bz: bugzoo.BugZoo,
                  bug: Bug,
                  filename: str
                  ) -> 'DonorPool':
        """
        Constructs a donor pool of snippets from the contents of a given file.
        """
        snippets = set()
        src_file = SourceFile.load(bz, bug, filename)

        # create a snippet for each line
        for content in src_file:
            content = content.strip()

            # skip comments
            if content.startswith('/*') or content.startswith('/'):
                continue

            # skip macros
            if content.startswith('#'):
                continue

            # restrict to statements
            if not content.endswith(';'):
                continue

            snippets.add(Snippet(content))

        return DonorPool(snippets)

    def __init__(self, snippets: Iterable[Snippet]):
        self.__snippets = frozenset(snippets)

    def __iter__(self) -> Iterator[Snippet]:
        """
        Returns an iterator over the snippets contained in this donor pool.
        """
        return self.__snippets.__iter__()
